@@ -5,12 +5,12 @@
 */
 #include <stdlib.h>
 #include <string.h>
-#include <assert.h>
 #ifdef _OS_WINDOWS_
 #include <malloc.h>
 #endif
 #include "julia.h"
 #include "julia_internal.h"
+#include "julia_assert.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -21,20 +21,19 @@ extern "C" {
 #define JL_ARRAY_ALIGN(jl_value, nbytes) LLT_ALIGN(jl_value, nbytes)
 
 // array constructors ---------------------------------------------------------
-
-static inline int store_unboxed(jl_value_t *el_type) // jl_isbits
+char *jl_array_typetagdata(jl_array_t *a) JL_NOTSAFEPOINT
 {
-    return jl_is_leaf_type(el_type) && jl_is_immutable(el_type) &&
-        ((jl_datatype_t*)el_type)->layout &&
-        ((jl_datatype_t*)el_type)->layout->npointers == 0;
+    assert(jl_array_isbitsunion(a));
+    return ((char*)jl_array_data(a)) + ((jl_array_ndims(a) == 1 ? (a->maxsize - a->offset) : jl_array_len(a)) * a->elsize) + a->offset;
 }
 
-int jl_array_store_unboxed(jl_value_t *el_type)
+JL_DLLEXPORT int jl_array_store_unboxed(jl_value_t *eltype) JL_NOTSAFEPOINT
 {
-    return store_unboxed(el_type);
+    size_t fsz = 0, al = 0;
+    return jl_islayout_inline(eltype, &fsz, &al);
 }
 
-STATIC_INLINE jl_value_t *jl_array_owner(jl_array_t *a)
+STATIC_INLINE jl_value_t *jl_array_owner(jl_array_t *a JL_PROPAGATES_ROOT) JL_NOTSAFEPOINT
 {
     if (a->flags.how == 3) {
         a = (jl_array_t*)jl_array_data_owner(a);
@@ -54,7 +53,7 @@ size_t jl_arr_xtralloc_limit = 0;
 #define MAXINTVAL (((size_t)-1)>>1)
 
 static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
-                               int isunboxed, int elsz)
+                               int isunboxed, int isunion, int elsz)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     size_t i, tot, nel=1;
@@ -62,21 +61,26 @@ static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
     jl_array_t *a;
 
     for(i=0; i < ndims; i++) {
-        wideint_t prod = (wideint_t)nel * (wideint_t)dims[i];
-        if (prod > (wideint_t) MAXINTVAL)
+        size_t di = dims[i];
+        wideint_t prod = (wideint_t)nel * (wideint_t)di;
+        if (prod > (wideint_t) MAXINTVAL || di > MAXINTVAL)
             jl_error("invalid Array dimensions");
         nel = prod;
     }
-
+    assert(atype == NULL || isunion == jl_is_uniontype(jl_tparam0(atype)));
     if (isunboxed) {
         wideint_t prod = (wideint_t)elsz * (wideint_t)nel;
         if (prod > (wideint_t) MAXINTVAL)
             jl_error("invalid Array size");
         tot = prod;
-        if (elsz == 1) {
+        if (elsz == 1 && !isunion) {
             // extra byte for all julia allocated byte arrays
             tot++;
         }
+        if (isunion) {
+            // an extra byte for each isbits union array element, stored after a->maxsize
+            tot += nel;
+        }
     }
     else {
         wideint_t prod = (wideint_t)sizeof(void*) * (wideint_t)nel;
@@ -97,7 +101,7 @@ static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
         // No allocation or safepoint allowed after this
         a->flags.how = 0;
         data = (char*)a + doffs;
-        if (tot > 0 && !isunboxed)
+        if ((tot > 0 && !isunboxed) || isunion)
             memset(data, 0, tot);
     }
     else {
@@ -109,7 +113,8 @@ static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
         // No allocation or safepoint allowed after this
         a->flags.how = 2;
         jl_gc_track_malloced_array(ptls, a);
-        if (!isunboxed)
+        if (!isunboxed || isunion)
+            // need to zero out isbits union array selector bytes to ensure a valid type index
             memset(data, 0, tot);
     }
     a->flags.pooled = tsz <= GC_MAX_SZCLASS;
@@ -141,21 +146,27 @@ static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
 
 static inline jl_array_t *_new_array(jl_value_t *atype, uint32_t ndims, size_t *dims)
 {
-    int isunboxed=0, elsz=sizeof(void*);
-    jl_value_t *el_type = jl_tparam0(atype);
-    isunboxed = store_unboxed(el_type);
-    if (isunboxed)
-        elsz = jl_datatype_size(el_type);
-    return _new_array_(atype, ndims, dims, isunboxed, elsz);
+    jl_value_t *eltype = jl_tparam0(atype);
+    size_t elsz = 0, al = 0;
+    if (!jl_is_kind(jl_typeof(eltype)))
+        jl_type_error_rt("Array", "element type", (jl_value_t*)jl_type_type, eltype);
+    int isunboxed = jl_islayout_inline(eltype, &elsz, &al);
+    int isunion = jl_is_uniontype(eltype);
+    if (!isunboxed) {
+        elsz = sizeof(void*);
+        al = elsz;
+    }
+
+    return _new_array_(atype, ndims, dims, isunboxed, isunion, elsz);
 }
 
 jl_array_t *jl_new_array_for_deserialization(jl_value_t *atype, uint32_t ndims, size_t *dims,
-                                             int isunboxed, int elsz)
+                                             int isunboxed, int isunion, int elsz)
 {
-    return _new_array_(atype, ndims, dims, isunboxed, elsz);
+    return _new_array_(atype, ndims, dims, isunboxed, isunion, elsz);
 }
 
-#ifndef NDEBUG
+#ifndef JL_NDEBUG
 static inline int is_ntuple_long(jl_value_t *v)
 {
     if (!jl_is_tuple(v))
@@ -178,9 +189,10 @@ JL_DLLEXPORT jl_array_t *jl_reshape_array(jl_value_t *atype, jl_array_t *data,
     size_t ndims = jl_nfields(_dims);
     assert(is_ntuple_long(_dims));
     size_t *dims = (size_t*)_dims;
+    assert(jl_types_equal(jl_tparam0(jl_typeof(data)), jl_tparam0(atype)));
 
     int ndimwords = jl_array_ndimwords(ndims);
-    int tsz = JL_ARRAY_ALIGN(sizeof(jl_array_t) + ndimwords*sizeof(size_t) + sizeof(void*), JL_SMALL_BYTE_ALIGNMENT);
+    int tsz = JL_ARRAY_ALIGN(sizeof(jl_array_t) + ndimwords * sizeof(size_t) + sizeof(void*), JL_SMALL_BYTE_ALIGNMENT);
     a = (jl_array_t*)jl_gc_alloc(ptls, tsz, atype);
     // No allocation or safepoint allowed after this
     a->flags.pooled = tsz <= GC_MAX_SZCLASS;
@@ -188,10 +200,25 @@ JL_DLLEXPORT jl_array_t *jl_reshape_array(jl_value_t *atype, jl_array_t *data,
     a->offset = 0;
     a->data = NULL;
     a->flags.isaligned = data->flags.isaligned;
-    jl_value_t *el_type = jl_tparam0(atype);
-    assert(store_unboxed(el_type) == !data->flags.ptrarray);
-    if (!data->flags.ptrarray) {
-        a->elsize = jl_datatype_size(el_type);
+    jl_array_t *owner = (jl_array_t*)jl_array_owner(data);
+    jl_value_t *eltype = jl_tparam0(atype);
+    size_t elsz = 0, align = 0;
+    int isboxed = !jl_islayout_inline(eltype, &elsz, &align);
+    assert(isboxed == data->flags.ptrarray);
+    if (!isboxed) {
+        a->elsize = elsz;
+        jl_value_t *ownerty = jl_typeof(owner);
+        size_t oldelsz = 0, oldalign = 0;
+        if (ownerty == (jl_value_t*)jl_string_type) {
+            oldalign = 1;
+        }
+        else {
+            jl_islayout_inline(jl_tparam0(ownerty), &oldelsz, &oldalign);
+        }
+        if (oldalign < align)
+            jl_exceptionf(jl_argumenterror_type,
+                          "reinterpret from alignment %d bytes to alignment %d bytes not allowed",
+                          (int) oldalign, (int) align);
         a->flags.ptrarray = 0;
     }
     else {
@@ -201,7 +228,7 @@ JL_DLLEXPORT jl_array_t *jl_reshape_array(jl_value_t *atype, jl_array_t *data,
 
     // if data is itself a shared wrapper,
     // owner should point back to the original array
-    jl_array_data_owner(a) = jl_array_owner(data);
+    jl_array_data_owner(a) = (jl_value_t*)owner;
 
     a->flags.how = 3;
     a->data = data->data;
@@ -266,15 +293,25 @@ JL_DLLEXPORT jl_array_t *jl_ptr_to_array_1d(jl_value_t *atype, void *data,
                                             size_t nel, int own_buffer)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
-    size_t elsz;
     jl_array_t *a;
-    jl_value_t *el_type = jl_tparam0(atype);
+    jl_value_t *eltype = jl_tparam0(atype);
 
-    int isunboxed = store_unboxed(el_type);
-    if (isunboxed)
-        elsz = jl_datatype_size(el_type);
-    else
-        elsz = sizeof(void*);
+    int isunboxed = jl_array_store_unboxed(eltype);
+    size_t elsz;
+    unsigned align;
+    if (isunboxed && jl_is_uniontype(eltype))
+        jl_exceptionf(jl_argumenterror_type,
+                      "unsafe_wrap: unspecified layout for union element type");
+    if (isunboxed) {
+        elsz = jl_datatype_size(eltype);
+        align = jl_datatype_align(eltype);
+    }
+    else {
+        align = elsz = sizeof(void*);
+    }
+    if (((uintptr_t)data) & (align - 1))
+        jl_exceptionf(jl_argumenterror_type,
+                      "unsafe_wrap: pointer %p is not properly aligned to %u bytes", data, align);
 
     int ndimwords = jl_array_ndimwords(1);
     int tsz = JL_ARRAY_ALIGN(sizeof(jl_array_t) + ndimwords*sizeof(size_t), JL_CACHE_BYTE_ALIGNMENT);
@@ -309,7 +346,7 @@ JL_DLLEXPORT jl_array_t *jl_ptr_to_array(jl_value_t *atype, void *data,
                                          jl_value_t *_dims, int own_buffer)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
-    size_t elsz, nel = 1;
+    size_t nel = 1;
     jl_array_t *a;
     size_t ndims = jl_nfields(_dims);
     wideint_t prod;
@@ -323,13 +360,24 @@ JL_DLLEXPORT jl_array_t *jl_ptr_to_array(jl_value_t *atype, void *data,
     }
     if (__unlikely(ndims == 1))
         return jl_ptr_to_array_1d(atype, data, nel, own_buffer);
-    jl_value_t *el_type = jl_tparam0(atype);
+    jl_value_t *eltype = jl_tparam0(atype);
 
-    int isunboxed = store_unboxed(el_type);
-    if (isunboxed)
-        elsz = jl_datatype_size(el_type);
-    else
-        elsz = sizeof(void*);
+    int isunboxed = jl_array_store_unboxed(eltype);
+    size_t elsz;
+    unsigned align;
+    if (isunboxed && jl_is_uniontype(eltype))
+        jl_exceptionf(jl_argumenterror_type,
+                      "unsafe_wrap: unspecified layout for union element type");
+    if (isunboxed) {
+        elsz = jl_datatype_size(eltype);
+        align = jl_datatype_align(eltype);
+    }
+    else {
+        align = elsz = sizeof(void*);
+    }
+    if (((uintptr_t)data) & (align - 1))
+        jl_exceptionf(jl_argumenterror_type,
+                      "unsafe_wrap: pointer %p is not properly aligned to %u bytes", data, align);
 
     int ndimwords = jl_array_ndimwords(ndims);
     int tsz = JL_ARRAY_ALIGN(sizeof(jl_array_t) + ndimwords*sizeof(size_t), JL_CACHE_BYTE_ALIGNMENT);
@@ -395,22 +443,34 @@ JL_DLLEXPORT jl_array_t *jl_pchar_to_array(const char *str, size_t len)
 
 JL_DLLEXPORT jl_value_t *jl_array_to_string(jl_array_t *a)
 {
+    size_t len = jl_array_len(a);
     if (a->flags.how == 3 && a->offset == 0 && a->elsize == 1 &&
         (jl_array_ndims(a) != 1 ||
-         !(a->maxsize+sizeof(void*)+1 > GC_MAX_SZCLASS && jl_array_nrows(a)+sizeof(void*)+1 <= GC_MAX_SZCLASS))) {
+         ((a->maxsize + sizeof(void*) + 1 <= GC_MAX_SZCLASS) == (len + sizeof(void*) + 1 <= GC_MAX_SZCLASS)))) {
         jl_value_t *o = jl_array_data_owner(a);
         if (jl_is_string(o)) {
             a->flags.isshared = 1;
-            *(size_t*)o = jl_array_len(a);
+            *(size_t*)o = len;
+            a->nrows = 0;
+#ifdef STORE_ARRAY_LEN
+            a->length = 0;
+#endif
+            a->maxsize = 0;
             return o;
         }
     }
-    return jl_pchar_to_string((const char*)jl_array_data(a), jl_array_len(a));
+    a->nrows = 0;
+#ifdef STORE_ARRAY_LEN
+    a->length = 0;
+#endif
+    a->maxsize = 0;
+    return jl_pchar_to_string((const char*)jl_array_data(a), len);
 }
 
 JL_DLLEXPORT jl_value_t *jl_pchar_to_string(const char *str, size_t len)
 {
-    jl_value_t *s = jl_gc_alloc(jl_get_ptls_states(), sizeof(size_t)+len+1, jl_string_type);
+    size_t sz = sizeof(size_t) + len + 1; // add space for trailing \nul protector and size
+    jl_value_t *s = jl_gc_alloc_(jl_get_ptls_states(), sz, jl_string_type); // force inlining
     *(size_t*)s = len;
     memcpy((char*)s + sizeof(size_t), str, len);
     ((char*)s + sizeof(size_t))[len] = 0;
@@ -419,7 +479,8 @@ JL_DLLEXPORT jl_value_t *jl_pchar_to_string(const char *str, size_t len)
 
 JL_DLLEXPORT jl_value_t *jl_alloc_string(size_t len)
 {
-    jl_value_t *s = jl_gc_alloc(jl_get_ptls_states(), sizeof(size_t)+len+1, jl_string_type);
+    size_t sz = sizeof(size_t) + len + 1; // add space for trailing \nul protector and size
+    jl_value_t *s = jl_gc_alloc_(jl_get_ptls_states(), sz, jl_string_type); // force inlining
     *(size_t*)s = len;
     ((char*)s + sizeof(size_t))[len] = 0;
     return s;
@@ -456,23 +517,34 @@ JL_DLLEXPORT size_t jl_array_len_(jl_array_t *a)
 }
 #endif
 
-JL_DLLEXPORT jl_value_t *jl_arrayref(jl_array_t *a, size_t i)
+JL_DLLEXPORT jl_value_t *jl_ptrarrayref(jl_array_t *a JL_PROPAGATES_ROOT, size_t i) JL_NOTSAFEPOINT
 {
     assert(i < jl_array_len(a));
-    jl_value_t *elt;
-    if (!a->flags.ptrarray) {
-        jl_value_t *el_type = (jl_value_t*)jl_tparam0(jl_typeof(a));
-        elt = jl_new_bits(el_type, &((char*)a->data)[i*a->elsize]);
-    }
-    else {
-        elt = ((jl_value_t**)a->data)[i];
-        if (elt == NULL) {
-            jl_throw(jl_undefref_exception);
-        }
+    assert(a->flags.ptrarray);
+    jl_value_t *elt = ((jl_value_t**)a->data)[i];
+    if (elt == NULL) {
+        jl_throw(jl_undefref_exception);
     }
     return elt;
 }
 
+
+JL_DLLEXPORT jl_value_t *jl_arrayref(jl_array_t *a, size_t i)
+{
+    if (a->flags.ptrarray)
+        return jl_ptrarrayref(a, i);
+    assert(i < jl_array_len(a));
+    jl_value_t *eltype = (jl_value_t*)jl_tparam0(jl_typeof(a));
+    if (jl_is_uniontype(eltype)) {
+        // isbits union selector bytes are always stored directly after the last array element
+        uint8_t sel = jl_array_typetagdata(a)[i];
+        eltype = jl_nth_union_component(eltype, sel);
+        if (jl_is_datatype_singleton((jl_datatype_t*)eltype))
+            return ((jl_datatype_t*)eltype)->instance;
+    }
+    return jl_new_bits(eltype, &((char*)a->data)[i * a->elsize]);
+}
+
 JL_DLLEXPORT int jl_array_isassigned(jl_array_t *a, size_t i)
 {
     if (a->flags.ptrarray)
@@ -480,48 +552,27 @@ JL_DLLEXPORT int jl_array_isassigned(jl_array_t *a, size_t i)
     return 1;
 }
 
-int jl_array_isdefined(jl_value_t **args0, int nargs)
-{
-    assert(jl_is_array(args0[0]));
-    jl_depwarn("`isdefined(a::Array, i::Int)` is deprecated, "
-               "use `isassigned(a, i)` instead", (jl_value_t*)jl_symbol("isdefined"));
-
-    jl_array_t *a = (jl_array_t*)args0[0];
-    jl_value_t **args = &args0[1];
-    size_t nidxs = nargs-1;
-    size_t i=0;
-    size_t k, stride=1;
-    size_t nd = jl_array_ndims(a);
-    for(k=0; k < nidxs; k++) {
-        if (!jl_is_long(args[k]))
-            jl_type_error("isdefined", (jl_value_t*)jl_long_type, args[k]);
-        size_t ii = jl_unbox_long(args[k])-1;
-        i += ii * stride;
-        size_t d = k>=nd ? 1 : jl_array_dim(a, k);
-        if (k < nidxs-1 && ii >= d)
-            return 0;
-        stride *= d;
-    }
-    for(; k < nd; k++)
-        stride *= jl_array_dim(a, k);
-    if (i >= stride)
-        return 0;
-
-    if (a->flags.ptrarray)
-        return ((jl_value_t**)jl_array_data(a))[i] != NULL;
-    return 1;
-}
-
-JL_DLLEXPORT void jl_arrayset(jl_array_t *a, jl_value_t *rhs, size_t i)
+JL_DLLEXPORT void jl_arrayset(jl_array_t *a JL_ROOTING_ARGUMENT, jl_value_t *rhs JL_ROOTED_ARGUMENT JL_MAYBE_UNROOTED, size_t i)
 {
     assert(i < jl_array_len(a));
-    jl_value_t *el_type = jl_tparam0(jl_typeof(a));
-    if (el_type != (jl_value_t*)jl_any_type) {
-        if (!jl_isa(rhs, el_type))
-            jl_type_error("arrayset", el_type, rhs);
+    jl_value_t *eltype = jl_tparam0(jl_typeof(a));
+    if (eltype != (jl_value_t*)jl_any_type) {
+        JL_GC_PUSH1(&rhs);
+        if (!jl_isa(rhs, eltype))
+            jl_type_error("arrayset", eltype, rhs);
+        JL_GC_POP();
     }
     if (!a->flags.ptrarray) {
-        jl_assign_bits(&((char*)a->data)[i*a->elsize], rhs);
+        if (jl_is_uniontype(eltype)) {
+            uint8_t *psel = &((uint8_t*)jl_array_typetagdata(a))[i];
+            unsigned nth = 0;
+            if (!jl_find_union_component(eltype, jl_typeof(rhs), &nth))
+                assert(0 && "invalid arrayset to isbits union");
+            *psel = nth;
+            if (jl_is_datatype_singleton((jl_datatype_t*)jl_typeof(rhs)))
+                return;
+        }
+        jl_assign_bits(&((char*)a->data)[i * a->elsize], rhs);
     }
     else {
         ((jl_value_t**)a->data)[i] = rhs;
@@ -557,11 +608,16 @@ static int NOINLINE array_resize_buffer(jl_array_t *a, size_t newlen)
     size_t oldnbytes = a->maxsize * elsz;
     size_t oldoffsnb = a->offset * elsz;
     size_t oldlen = a->nrows;
+    int isbitsunion = jl_array_isbitsunion(a);
     assert(nbytes >= oldnbytes);
-    if (elsz == 1) {
+    if (elsz == 1 && !isbitsunion) {
         nbytes++;
         oldnbytes++;
     }
+    if (isbitsunion) {
+        nbytes += newlen;
+        oldnbytes += a->maxsize;
+    }
     int newbuf = 0;
     if (a->flags.how == 2) {
         // already malloc'd - use realloc
@@ -569,15 +625,15 @@ static int NOINLINE array_resize_buffer(jl_array_t *a, size_t newlen)
         a->data = jl_gc_managed_realloc(olddata, nbytes, oldnbytes,
                                         a->flags.isaligned, (jl_value_t*)a);
     }
-    else if (a->flags.how == 3 && jl_is_string(jl_array_data_owner(a))) {
+    else if (a->flags.how == 3 && jl_is_string(jl_array_data_owner(a)) && !isbitsunion) {
         // if data is in a String, keep it that way
         jl_value_t *s;
         if (a->flags.isshared) {
-            s = jl_alloc_string(nbytes);
+            s = jl_alloc_string(nbytes - (elsz == 1));
             newbuf = 1;
         }
         else {
-            s = jl_gc_realloc_string(jl_array_data_owner(a), nbytes);
+            s = jl_gc_realloc_string(jl_array_data_owner(a), nbytes - (elsz == 1));
         }
         jl_array_data_owner(a) = s;
         jl_gc_wb(a, s);
@@ -603,7 +659,7 @@ static int NOINLINE array_resize_buffer(jl_array_t *a, size_t newlen)
             jl_gc_wb_buf(a, a->data, nbytes);
         }
     }
-    if (JL_ARRAY_IMPL_NUL && elsz == 1)
+    if (JL_ARRAY_IMPL_NUL && elsz == 1 && !isbitsunion)
         memset((char*)a->data + oldnbytes - 1, 0, nbytes - oldnbytes + 1);
     (void)oldlen;
     assert(oldlen == a->nrows &&
@@ -622,9 +678,11 @@ static void NOINLINE array_try_unshare(jl_array_t *a)
         if (jl_is_string(jl_array_data_owner(a)))
             return;
         assert(a->offset == 0);
-        size_t len = jl_array_nrows(a);
-        size_t es = a->elsize;
-        size_t nbytes = len * es;
+        size_t len = a->maxsize;
+        size_t nbytes = len * a->elsize;
+        if (jl_array_isbitsunion(a)) {
+            nbytes += len;
+        }
         char *olddata = (char*)a->data;
         int newbuf = array_resize_buffer(a, len);
         assert(newbuf);
@@ -664,44 +722,80 @@ STATIC_INLINE void jl_array_grow_at_beg(jl_array_t *a, size_t idx, size_t inc,
     size_t nbinc = inc * elsz;
     char *data = (char*)a->data;
     char *newdata;
+    char *typetagdata;
+    char *newtypetagdata;
+    int isbitsunion = jl_array_isbitsunion(a);
+    if (isbitsunion) typetagdata = jl_array_typetagdata(a);
     if (a->offset >= inc) {
+        // already have enough space in a->offset
         newdata = data - nbinc;
         a->offset -= inc;
+        if (isbitsunion) newtypetagdata = typetagdata - inc;
         if (idx > 0) {
+            // inserting new elements after 1st element
             memmove(newdata, data, idx * elsz);
+            if (isbitsunion) {
+                memmove(newtypetagdata, typetagdata, idx);
+                memset(newtypetagdata + idx, 0, inc);
+            }
         }
     }
     else {
-        size_t oldoffsnb = a->offset * elsz;
+        // not enough room for requested growth from existing a->offset
+        size_t oldoffset = a->offset;
+        size_t oldoffsnb = oldoffset * elsz;
+        size_t oldmaxsize = a->maxsize;
         size_t nb1 = idx * elsz;
         if (inc > (a->maxsize - n) / 2 - (a->maxsize - n) / 20) {
+            // not enough room for requested growth from end of array
             size_t newlen = a->maxsize == 0 ? inc * 2 : a->maxsize * 2;
             while (n + 2 * inc > newlen - a->offset)
                 newlen *= 2;
             newlen = limit_overallocation(a, n, newlen, 2 * inc);
             size_t newoffset = (newlen - newnrows) / 2;
-            if (!array_resize_buffer(a, newlen))
+            if (!array_resize_buffer(a, newlen)) {
                 data = (char*)a->data + oldoffsnb;
+            }
             newdata = (char*)a->data + newoffset * elsz;
+            if (isbitsunion) {
+                typetagdata = data + (oldmaxsize - oldoffset) * elsz + oldoffset;
+                newtypetagdata = newdata + (a->maxsize - newoffset) * elsz + newoffset;
+                memmove(newtypetagdata, typetagdata, idx);
+                memset(newtypetagdata + idx, 0, inc);
+                memmove(newtypetagdata + idx + inc, typetagdata + idx, n - idx);
+            }
             // We could use memcpy if resizing allocates a new buffer,
             // hopefully it's not a particularly important optimization.
-            if (idx > 0 && newdata < data)
+            if (idx > 0 && newdata < data) {
                 memmove(newdata, data, nb1);
+            }
             memmove(newdata + nbinc + nb1, data + nb1, n * elsz - nb1);
-            if (idx > 0 && newdata > data)
+            if (idx > 0 && newdata > data) {
                 memmove(newdata, data, nb1);
+            }
             a->offset = newoffset;
         }
         else {
+            // use extra space between a->nrows & a->maxsize
             a->offset = (a->maxsize - newnrows) / 2;
             newdata = data - oldoffsnb + a->offset * elsz;
-            // We could use memcpy if resizing allocates a new buffer,
-            // hopefully it's not a particularly important optimization.
-            if (idx > 0 && newdata < data)
+            if (isbitsunion) newtypetagdata = newdata + (a->maxsize - a->offset) * elsz + a->offset;
+            if (idx > 0 && newdata < data) {
                 memmove(newdata, data, nb1);
+                if (isbitsunion) {
+                    memmove(newtypetagdata, typetagdata, idx);
+                    memset(newtypetagdata + idx, 0, inc);
+                }
+            }
             memmove(newdata + nbinc + nb1, data + nb1, n * elsz - nb1);
-            if (idx > 0 && newdata > data)
+            if (isbitsunion) memmove(newtypetagdata + idx + inc, typetagdata + idx, n - idx);
+            if (idx > 0 && newdata > data) {
                 memmove(newdata, data, nb1);
+                if (isbitsunion) {
+                    memmove(newtypetagdata, typetagdata, idx);
+                    memset(newtypetagdata + idx, 0, inc);
+                }
+            }
         }
     }
 #ifdef STORE_ARRAY_LEN
@@ -712,6 +806,9 @@ STATIC_INLINE void jl_array_grow_at_beg(jl_array_t *a, size_t idx, size_t inc,
     if (a->flags.ptrarray) {
         memset(newdata + idx * elsz, 0, nbinc);
     }
+    else if (isbitsunion) {
+        memset(newtypetagdata + idx, 0, inc);
+    }
 }
 
 STATIC_INLINE void jl_array_grow_at_end(jl_array_t *a, size_t idx,
@@ -730,31 +827,58 @@ STATIC_INLINE void jl_array_grow_at_end(jl_array_t *a, size_t idx,
     }
     size_t elsz = a->elsize;
     char *data = (char*)a->data;
+    char *typetagdata;
+    char *newtypetagdata;
+    int isbitsunion = jl_array_isbitsunion(a);
+    if (isbitsunion) typetagdata = jl_array_typetagdata(a);
     int has_gap = n > idx;
-    if (__unlikely((n + inc) > a->maxsize - a->offset)) {
+    size_t reqmaxsize = a->offset + n + inc;
+    if (__unlikely(reqmaxsize > a->maxsize)) {
         size_t nb1 = idx * elsz;
         size_t nbinc = inc * elsz;
-        size_t newlen = a->maxsize == 0 ? (inc < 4 ? 4 : inc) : a->maxsize * 2;
-        while ((n + inc) > newlen - a->offset)
-            newlen *= 2;
-        newlen = limit_overallocation(a, n, newlen, inc);
-        int newbuf = array_resize_buffer(a, newlen);
+        // if the requested size is more than 2x current maxsize, grow exactly
+        // otherwise double the maxsize
+        size_t newmaxsize = reqmaxsize >= a->maxsize * 2
+                          ? (reqmaxsize < 4 ? 4 : reqmaxsize)
+                          : a->maxsize * 2;
+        newmaxsize = limit_overallocation(a, n, newmaxsize, inc);
+        size_t oldmaxsize = a->maxsize;
+        int newbuf = array_resize_buffer(a, newmaxsize);
         char *newdata = (char*)a->data + a->offset * elsz;
+        if (isbitsunion) newtypetagdata = newdata + (a->maxsize - a->offset) * elsz + a->offset;
         if (newbuf) {
             memcpy(newdata, data, nb1);
-            if (has_gap) {
-                memcpy(newdata + nb1 + nbinc, data + nb1, n * elsz - nb1);
+            if (isbitsunion) {
+                memcpy(newtypetagdata, typetagdata, idx);
+                if (has_gap) memcpy(newtypetagdata + idx + inc, typetagdata + idx, n - idx);
+                memset(newtypetagdata + idx, 0, inc);
             }
+            if (has_gap) memcpy(newdata + nb1 + nbinc, data + nb1, n * elsz - nb1);
         }
-        else if (has_gap) {
-            memmove(newdata + nb1 + nbinc, newdata + nb1, n * elsz - nb1);
+        else {
+            if (isbitsunion) {
+                typetagdata = newdata + (oldmaxsize - a->offset) * elsz + a->offset;
+                if (has_gap) memmove(newtypetagdata + idx + inc, typetagdata + idx, n - idx);
+                memmove(newtypetagdata, typetagdata, idx);
+                memset(newtypetagdata + idx, 0, inc);
+            }
+            if (has_gap) memmove(newdata + nb1 + nbinc, newdata + nb1, n * elsz - nb1);
         }
         a->data = data = newdata;
     }
     else if (has_gap) {
+        if (isbitsunion) {
+            memmove(typetagdata + idx + inc, typetagdata + idx, n - idx);
+            memset(typetagdata + idx, 0, inc);
+        }
         size_t nb1 = idx * elsz;
         memmove(data + nb1 + inc * elsz, data + nb1, n * elsz - nb1);
     }
+    else {
+        // there was enough room for requested growth already in a->maxsize
+        if (isbitsunion)
+            memset(typetagdata + idx, 0, inc);
+    }
     size_t newnrows = n + inc;
 #ifdef STORE_ARRAY_LEN
     a->length = newnrows;
@@ -792,6 +916,51 @@ JL_DLLEXPORT void jl_array_grow_beg(jl_array_t *a, size_t inc)
     jl_array_grow_at_beg(a, 0, inc, n);
 }
 
+STATIC_INLINE void jl_array_shrink(jl_array_t *a, size_t dec)
+{
+    //if we don't manage this array return
+    if (a->flags.how == 0) return;
+
+    size_t elsz = a->elsize;
+    int newbytes = (a->maxsize - dec) * a->elsize;
+    int oldnbytes = (a->maxsize) * a->elsize;
+    int isbitsunion = jl_array_isbitsunion(a);
+    if (isbitsunion) {
+        newbytes += a->maxsize - dec;
+        oldnbytes += a->maxsize;
+    }
+
+    if (elsz == 1 && !isbitsunion) {
+        newbytes++;
+        oldnbytes++;
+    }
+    char *originalptr = ((char*) a->data) - a->offset * a->elsize;
+    if (a->flags.how == 1) {
+        //this is a julia-allocated buffer that needs to be marked
+    }
+    else if (a->flags.how == 2) {
+        //malloc-allocated pointer this array object manages
+        char *typetagdata;
+        char *newtypetagdata;
+        if (isbitsunion) {
+            typetagdata = (char*)malloc(a->nrows);
+            memcpy(typetagdata, jl_array_typetagdata(a), a->nrows);
+        }
+        size_t oldoffsnb = a->offset * elsz;
+        a->data = ((char*) jl_gc_managed_realloc(originalptr, newbytes, oldnbytes,
+                                        a->flags.isaligned, (jl_value_t*) a)) + oldoffsnb;
+        a->maxsize -= dec;
+        if (isbitsunion) {
+            newtypetagdata = jl_array_typetagdata(a);
+            memcpy(newtypetagdata, typetagdata, a->nrows);
+            free(typetagdata);
+        }
+    }
+    else if (a->flags.how == 3) {
+        //this has has a pointer to the object that owns the data
+    }
+}
+
 static size_t jl_array_limit_offset(jl_array_t *a, size_t offset)
 {
     // make sure offset doesn't grow forever due to deleting at beginning
@@ -813,6 +982,7 @@ STATIC_INLINE void jl_array_del_at_beg(jl_array_t *a, size_t idx, size_t dec,
     // assume inbounds, assume unshared
     size_t elsz = a->elsize;
     size_t offset = a->offset;
+    int isbitsunion = jl_array_isbitsunion(a);
     offset += dec;
 #ifdef STORE_ARRAY_LEN
     a->length = n - dec;
@@ -824,20 +994,32 @@ STATIC_INLINE void jl_array_del_at_beg(jl_array_t *a, size_t idx, size_t dec,
     if (__unlikely(newoffs != offset) || idx > 0) {
         char *olddata = (char*)a->data;
         char *newdata = olddata - (a->offset - newoffs) * elsz;
+        char *typetagdata;
+        char *newtypetagdata;
+        if (isbitsunion) {
+            typetagdata = jl_array_typetagdata(a);
+            newtypetagdata = typetagdata - (a->offset - newoffs);
+        }
+
         size_t nb1 = idx * elsz; // size in bytes of the first block
         size_t nbtotal = a->nrows * elsz; // size in bytes of the new array
         // Implicit '\0' for byte arrays
-        if (elsz == 1)
+        if (elsz == 1 && !isbitsunion)
             nbtotal++;
-        if (idx > 0)
+        if (idx > 0) {
             memmove(newdata, olddata, nb1);
+            if (isbitsunion) memmove(newtypetagdata, typetagdata, idx);
+        }
         // Move the rest of the data if the offset changed
-        if (newoffs != offset)
+        if (newoffs != offset) {
             memmove(newdata + nb1, olddata + nb1 + nbdec, nbtotal - nb1);
+            if (isbitsunion) memmove(newtypetagdata + idx, typetagdata + idx + dec, n - idx);
+        }
         a->data = newdata;
     }
     else {
-        a->data = (char*)a->data + nbdec;
+        char *data = (char*)a->data;
+        a->data = data + nbdec;
     }
     a->offset = newoffs;
 }
@@ -849,11 +1031,17 @@ STATIC_INLINE void jl_array_del_at_end(jl_array_t *a, size_t idx, size_t dec,
     // assume inbounds, assume unshared
     char *data = (char*)a->data;
     size_t elsz = a->elsize;
+    int isbitsunion = jl_array_isbitsunion(a);
     size_t last = idx + dec;
-    if (n > last)
+    if (n > last) {
         memmove(data + idx * elsz, data + last * elsz, (n - last) * elsz);
+        if (isbitsunion) {
+            char *typetagdata = jl_array_typetagdata(a);
+            memmove(typetagdata + idx, typetagdata + last, n - last);
+        }
+    }
     n -= dec;
-    if (elsz == 1)
+    if (elsz == 1 && !isbitsunion)
         data[n] = 0;
     a->nrows = n;
 #ifdef STORE_ARRAY_LEN
@@ -887,6 +1075,8 @@ JL_DLLEXPORT void jl_array_del_beg(jl_array_t *a, size_t dec)
         jl_bounds_error_int((jl_value_t*)a, dec);
     if (__unlikely(a->flags.isshared))
         array_try_unshare(a);
+    if (dec == 0)
+        return;
     jl_array_del_at_beg(a, 0, dec, n);
 }
 
@@ -897,28 +1087,47 @@ JL_DLLEXPORT void jl_array_del_end(jl_array_t *a, size_t dec)
         jl_bounds_error_int((jl_value_t*)a, 0);
     if (__unlikely(a->flags.isshared))
         array_try_unshare(a);
+    if (dec == 0)
+        return;
     jl_array_del_at_end(a, n - dec, dec, n);
 }
 
 JL_DLLEXPORT void jl_array_sizehint(jl_array_t *a, size_t sz)
 {
     size_t n = jl_array_nrows(a);
-    if (sz <= n)
-        return;
-    size_t inc = sz - n;
-    jl_array_grow_end(a, inc);
-    a->nrows = n;
+
+    int min = a->offset + a->length;
+    sz = (sz < min) ? min : sz;
+
+    if (sz <= a->maxsize) {
+        size_t dec = a->maxsize - sz;
+        //if we don't save at least an eighth of maxsize then its not worth it to shrink
+        if (dec < a->maxsize / 8) return;
+        jl_array_shrink(a, dec);
+    }
+    else {
+        size_t inc = sz - n;
+        jl_array_grow_end(a, inc);
+
+        a->nrows = n;
 #ifdef STORE_ARRAY_LEN
-    a->length = n;
+        a->length = n;
 #endif
+    }
 }
 
 JL_DLLEXPORT jl_array_t *jl_array_copy(jl_array_t *ary)
 {
     size_t elsz = ary->elsize;
+    size_t len = jl_array_len(ary);
+    int isunion = jl_is_uniontype(jl_tparam0(jl_typeof(ary)));
     jl_array_t *new_ary = _new_array_(jl_typeof(ary), jl_array_ndims(ary),
-                                      &ary->nrows, !ary->flags.ptrarray, elsz);
-    memcpy(new_ary->data, ary->data, jl_array_len(ary) * elsz);
+                                      &ary->nrows, !ary->flags.ptrarray,
+                                      isunion, elsz);
+    memcpy(new_ary->data, ary->data, len * elsz);
+    // ensure isbits union arrays copy their selector bytes correctly
+    if (jl_array_isbitsunion(ary))
+        memcpy(jl_array_typetagdata(new_ary), jl_array_typetagdata(ary), len);
     return new_ary;
 }
 
@@ -1004,15 +1213,6 @@ JL_DLLEXPORT void jl_array_ptr_1d_append(jl_array_t *a, jl_array_t *a2)
     }
 }
 
-JL_DLLEXPORT void jl_array_ptr_1d_push2(jl_array_t *a, jl_value_t *b, jl_value_t *c)
-{
-    assert(jl_typeis(a, jl_array_any_type));
-    jl_array_grow_end(a, 2);
-    size_t n = jl_array_nrows(a);
-    jl_array_ptr_set(a, n - 2, b);
-    jl_array_ptr_set(a, n - 1, c);
-}
-
 JL_DLLEXPORT jl_value_t *(jl_array_data_owner)(jl_array_t *a)
 {
     return jl_array_data_owner(a);
