@@ -10,6 +10,7 @@
 #else
 #include "getopt.h"
 #endif
+#include "julia_assert.h"
 
 #ifdef _OS_WINDOWS_
 char *shlib_ext = ".dll";
@@ -19,30 +20,23 @@ char *shlib_ext = ".dylib";
 char *shlib_ext = ".so";
 #endif
 
-static char system_image_path[256] = "\0" JL_SYSTEM_IMAGE_PATH;
+static const char system_image_path[256] = "\0" JL_SYSTEM_IMAGE_PATH;
 JL_DLLEXPORT const char *jl_get_default_sysimg_path(void)
 {
-#ifdef CPUID_SPECIFIC_BINARIES
-    char *path = &system_image_path[1];
-    size_t existing_length = strlen(path) - strlen(shlib_ext);
-    path += existing_length;
-    snprintf(path, sizeof(system_image_path) - existing_length,
-        "_%" PRIx64 "%s", jl_cpuid_tag(), shlib_ext);
-#endif
     return &system_image_path[1];
 }
 
 
 jl_options_t jl_options = { 0,    // quiet
-                            NULL, // julia_home
+                            -1,   // banner
+                            NULL, // julia_bindir
                             NULL, // julia_bin
-                            NULL, // eval
-                            NULL, // print
-                            NULL, // load
+                            NULL, // cmds
                             NULL, // image_file (will be filled in below)
                             NULL, // cpu_target ("native", "core2", etc...)
                             0,    // nprocs
-                            NULL, // machinefile
+                            NULL, // machine_file
+                            NULL, // project
                             0,    // isinteractive
                             0,    // color
                             JL_OPTIONS_HISTORYFILE_ON, // history file
@@ -57,18 +51,24 @@ jl_options_t jl_options = { 0,    // quiet
                             1,    // debug_level [release build]
 #endif
                             JL_OPTIONS_CHECK_BOUNDS_DEFAULT, // check_bounds
-                            1,    // deprecation warning
+                            JL_OPTIONS_DEPWARN_ON,    // deprecation warning
+                            0,    // method overwrite warning
                             1,    // can_inline
                             JL_OPTIONS_POLLY_ON, // polly
+                            NULL, // trace_compile
                             JL_OPTIONS_FAST_MATH_DEFAULT,
                             0,    // worker
+                            NULL, // cookie
                             JL_OPTIONS_HANDLE_SIGNALS_ON,
-                            JL_OPTIONS_USE_PRECOMPILED_YES,
-                            JL_OPTIONS_USE_COMPILECACHE_YES,
+                            JL_OPTIONS_USE_SYSIMAGE_NATIVE_CODE_YES,
+                            JL_OPTIONS_USE_COMPILED_MODULES_YES,
                             NULL, // bind-to
                             NULL, // output-bc
+                            NULL, // output-unopt-bc
+                            NULL, // output-jit-bc
                             NULL, // output-o
                             NULL, // output-ji
+                            NULL,    // output-code_coverage
                             0, // incremental
                             0 // image_file_specified
 };
@@ -79,32 +79,40 @@ static const char opts[]  =
     " -h, --help                Print this message\n\n"
 
     // startup options
+    " --project[={<dir>|@.}]    Set <dir> as the home project/environment\n"
     " -J, --sysimage <file>     Start up with the given system image file\n"
-    " --precompiled={yes|no}    Use precompiled code from system image if available\n"
-    " --compilecache={yes|no}   Enable/disable incremental precompilation of modules\n"
     " -H, --home <dir>          Set location of `julia` executable\n"
-    " --startup-file={yes|no}   Load ~/.juliarc.jl\n"
-    " --handle-signals={yes|no} Enable or disable Julia's default signal handlers\n\n"
+    " --startup-file={yes|no}   Load `~/.julia/config/startup.jl`\n"
+    " --handle-signals={yes|no} Enable or disable Julia's default signal handlers\n"
+    " --sysimage-native-code={yes|no}\n"
+    "                           Use native code from system image if available\n"
+    " --compiled-modules={yes|no}\n"
+    "                           Enable or disable incremental precompilation of modules\n\n"
 
     // actions
     " -e, --eval <expr>         Evaluate <expr>\n"
-    " -E, --print <expr>        Evaluate and show <expr>\n"
+    " -E, --print <expr>        Evaluate <expr> and display the result\n"
     " -L, --load <file>         Load <file> immediately on all processors\n\n"
 
     // parallel options
     " -p, --procs {N|auto}      Integer value N launches N additional local worker processes\n"
-    "                           \"auto\" launches as many workers as the number of local cores\n"
-    " --machinefile <file>      Run processes on hosts listed in <file>\n\n"
+    "                           \"auto\" launches as many workers as the number of local CPU threads (logical cores)\n"
+    " --machine-file <file>     Run processes on hosts listed in <file>\n\n"
 
     // interactive options
     " -i                        Interactive mode; REPL runs and isinteractive() is true\n"
-    " -q, --quiet               Quiet startup (no banner)\n"
-    " --color={yes|no}          Enable or disable color text\n"
+    " -q, --quiet               Quiet startup: no banner, suppress REPL warnings\n"
+    " --banner={yes|no|auto}    Enable or disable startup banner\n"
+    " --color={yes|no|auto}     Enable or disable color text\n"
     " --history-file={yes|no}   Load or save history\n\n"
 
+    // error and warning options
+    " --depwarn={yes|no|error}  Enable or disable syntax and method deprecation warnings (\"error\" turns warnings into errors)\n"
+    " --warn-overwrite={yes|no} Enable or disable method overwrite warnings\n\n"
+
     // code generation options
-    " --compile={yes|no|all|min}Enable or disable JIT compiler, or request exhaustive compilation\n"
-    " -C, --cpu-target <target> Limit usage of cpu features up to <target>; set to \"help\" to see the available options\n"
+    //" --compile={yes|no|all|min}Enable or disable JIT compiler, or request exhaustive compilation\n"
+    " -C, --cpu-target <target> Limit usage of CPU features up to <target>; set to \"help\" to see the available options\n"
     " -O, --optimize={0,1,2,3}  Set the optimization level (default level is 2 if unspecified or 3 if used without a level)\n"
     " -g, -g <level>            Enable / Set the level of debug info generation"
 #ifdef JL_DEBUG_BUILD
@@ -112,27 +120,32 @@ static const char opts[]  =
 #else
         " (default level is 1 if unspecified or 2 if used without a level)\n"
 #endif
-    " --inline={yes|no}         Control whether inlining is permitted (overrides functions declared as @inline)\n"
+    " --inline={yes|no}         Control whether inlining is permitted, including overriding @inline declarations\n"
     " --check-bounds={yes|no}   Emit bounds checks always or never (ignoring declarations)\n"
 #ifdef USE_POLLY
     " --polly={yes|no}          Enable or disable the polyhedral optimizer Polly (overrides @polly declaration)\n"
 #endif
     " --math-mode={ieee,fast}   Disallow or enable unsafe floating point optimizations (overrides @fastmath declaration)\n\n"
 
-    // error and warning options
-    " --depwarn={yes|no|error}  Enable or disable syntax and method deprecation warnings (\"error\" turns warnings into errors)\n\n"
-
-    // compiler output options
-    " --output-o name           Generate an object file (including system image data)\n"
-    " --output-ji name          Generate a system image data file (.ji)\n"
-    " --output-bc name          Generate LLVM bitcode (.bc)\n"
-    " --output-incremental=no   Generate an incremental output file (rather than complete)\n\n"
-
     // instrumentation options
     " --code-coverage={none|user|all}, --code-coverage\n"
     "                           Count executions of source lines (omitting setting is equivalent to \"user\")\n"
+    " --code-coverage=tracefile.info\n"
+    "                           Append coverage information to the LCOV tracefile (filename supports format tokens).\n"
+// TODO: These TOKENS are defined in `runtime_ccall.cpp`. A more verbose `--help` should include that list here.
     " --track-allocation={none|user|all}, --track-allocation\n"
     "                           Count bytes allocated by each source line\n\n"
+
+    // compiler output options
+    //" --output-o name           Generate an object file (including system image data)\n"
+    //" --output-ji name          Generate a system image data file (.ji)\n"
+// These are for compiler debugging purposes only and should not be otherwise
+// used, so don't show them here. See the devdocs for tips on using these
+// options for debugging the compiler.
+//  " --output-unopt-bc name    Generate unoptimized LLVM bitcode (.bc)\n"
+//  " --output-jit-bc name      Dump all IR generated by the frontend (not including system image)\n"
+    //" --output-bc name          Generate LLVM bitcode (.bc)\n"
+    //" --output-incremental=no   Generate an incremental output file (rather than complete)\n\n"
 ;
 
 JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
@@ -145,10 +158,14 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
            opt_code_coverage,
            opt_track_allocation,
            opt_check_bounds,
+           opt_output_jit_bc,
+           opt_output_unopt_bc,
            opt_output_bc,
            opt_depwarn,
+           opt_warn_overwrite,
            opt_inline,
            opt_polly,
+           opt_trace_compile,
            opt_math_mode,
            opt_worker,
            opt_bind_to,
@@ -157,7 +174,12 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
            opt_output_ji,
            opt_use_precompiled,
            opt_use_compilecache,
-           opt_incremental
+           opt_incremental,
+           opt_banner,
+           opt_sysimage_native_code,
+           opt_compiled_modules,
+           opt_machine_file,
+           opt_project,
     };
     static const char* const shortopts = "+vhqH:e:E:L:J:C:ip:O:g:";
     static const struct option longopts[] = {
@@ -167,16 +189,18 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
         { "version",         no_argument,       0, 'v' },
         { "help",            no_argument,       0, 'h' },
         { "quiet",           no_argument,       0, 'q' },
+        { "banner",          required_argument, 0, opt_banner },
         { "home",            required_argument, 0, 'H' },
         { "eval",            required_argument, 0, 'e' },
         { "print",           required_argument, 0, 'E' },
         { "load",            required_argument, 0, 'L' },
         { "sysimage",        required_argument, 0, 'J' },
-        { "precompiled",     required_argument, 0, opt_use_precompiled },
-        { "compilecache",    required_argument, 0, opt_use_compilecache },
+        { "sysimage-native-code", required_argument, 0, opt_sysimage_native_code },
+        { "compiled-modules",    required_argument, 0, opt_compiled_modules },
         { "cpu-target",      required_argument, 0, 'C' },
         { "procs",           required_argument, 0, 'p' },
-        { "machinefile",     required_argument, 0, opt_machinefile },
+        { "machine-file",    required_argument, 0, opt_machine_file },
+        { "project",         optional_argument, 0, opt_project },
         { "color",           required_argument, 0, opt_color },
         { "history-file",    required_argument, 0, opt_history_file },
         { "startup-file",    required_argument, 0, opt_startup_file },
@@ -186,16 +210,20 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
         { "optimize",        optional_argument, 0, 'O' },
         { "check-bounds",    required_argument, 0, opt_check_bounds },
         { "output-bc",       required_argument, 0, opt_output_bc },
+        { "output-unopt-bc", required_argument, 0, opt_output_unopt_bc },
+        { "output-jit-bc",   required_argument, 0, opt_output_jit_bc },
         { "output-o",        required_argument, 0, opt_output_o },
         { "output-ji",       required_argument, 0, opt_output_ji },
         { "output-incremental",required_argument, 0, opt_incremental },
         { "depwarn",         required_argument, 0, opt_depwarn },
+        { "warn-overwrite",  required_argument, 0, opt_warn_overwrite },
         { "inline",          required_argument, 0, opt_inline },
         { "polly",           required_argument, 0, opt_polly },
+        { "trace-compile",   required_argument, 0, opt_trace_compile },
         { "math-mode",       required_argument, 0, opt_math_mode },
         { "handle-signals",  required_argument, 0, opt_handle_signals },
         // hidden command line options
-        { "worker",          required_argument, 0, opt_worker },
+        { "worker",          optional_argument, 0, opt_worker },
         { "bind-to",         required_argument, 0, opt_bind_to },
         { "lisp",            no_argument,       0, 1 },
         { 0, 0, 0, 0 }
@@ -204,14 +232,17 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
     // If CPUID specific binaries are enabled, this varies between runs, so initialize
     // it here, rather than as part of the static initialization above.
     jl_options.image_file = jl_get_default_sysimg_path();
+    jl_options.cmds = NULL;
 
-    int codecov  = JL_LOG_NONE;
-    int malloclog= JL_LOG_NONE;
+    int ncmds = 0;
+    const char **cmds = NULL;
+    int codecov = JL_LOG_NONE;
+    int malloclog = JL_LOG_NONE;
     // getopt handles argument parsing up to -- delineator
     int argc = *argcp;
     char **argv = *argvp;
     if (argc > 0) {
-        for (int i=0; i < argc; i++) {
+        for (int i = 0; i < argc; i++) {
             if (!strcmp(argv[i], "--")) {
                 argc = i;
                 break;
@@ -264,9 +295,6 @@ restart_switch:
         case 'h': // help
             jl_printf(JL_STDOUT, "%s%s", usage, opts);
             jl_exit(0);
-        case 'q': // quiet
-            jl_options.quiet = 1;
-            break;
         case 'g': // debug info
             if (optarg != NULL) {
                 if (!strcmp(optarg,"0"))
@@ -284,44 +312,79 @@ restart_switch:
             }
             break;
         case 'H': // home
-            jl_options.julia_home = strdup(optarg);
+            jl_options.julia_bindir = strdup(optarg);
+            if (!jl_options.julia_bindir)
+                jl_errorf("fatal error: failed to allocate memory: %s", strerror(errno));
             break;
         case 'e': // eval
-            jl_options.eval = strdup(optarg);
-            break;
         case 'E': // print
-            jl_options.print = strdup(optarg);
-            break;
         case 'L': // load
-            jl_options.load = strdup(optarg);
+        {
+            size_t sz = strlen(optarg) + 1;
+            char *arg = (char*)malloc(sz + 1);
+            const char **newcmds;
+            if (!arg)
+                jl_errorf("fatal error: failed to allocate memory: %s", strerror(errno));
+            arg[0] = c;
+            memcpy(arg + 1, optarg, sz);
+            newcmds = (const char**)realloc(cmds, (ncmds + 2) * sizeof(char*));
+            if (!newcmds) {
+                free(cmds);
+                jl_errorf("fatal error: failed to allocate memory: %s", strerror(errno));
+            }
+            cmds = newcmds;
+            cmds[ncmds] = arg;
+            ncmds++;
+            cmds[ncmds] = 0;
+            jl_options.cmds = cmds;
             break;
+        }
         case 'J': // sysimage
             jl_options.image_file = strdup(optarg);
+            if (!jl_options.image_file)
+                jl_errorf("fatal error: failed to allocate memory: %s", strerror(errno));
             jl_options.image_file_specified = 1;
             break;
-        case opt_use_precompiled:
-            if (!strcmp(optarg,"yes"))
-                jl_options.use_precompiled = JL_OPTIONS_USE_PRECOMPILED_YES;
-            else if (!strcmp(optarg,"no"))
-                jl_options.use_precompiled = JL_OPTIONS_USE_PRECOMPILED_NO;
-            else
-                jl_errorf("julia: invalid argument to --precompiled={yes|no} (%s)", optarg);
+        case 'q': // quiet
+            jl_options.quiet = 1;
+            if (jl_options.banner < 0)
+                jl_options.banner = 0;
             break;
-        case opt_use_compilecache:
-            if (!strcmp(optarg,"yes"))
-                jl_options.use_compilecache = JL_OPTIONS_USE_COMPILECACHE_YES;
-            else if (!strcmp(optarg,"no"))
-                jl_options.use_compilecache = JL_OPTIONS_USE_COMPILECACHE_NO;
+        case opt_banner: // banner
+            if (!strcmp(optarg, "yes"))
+                jl_options.banner = 1;
+            else if (!strcmp(optarg, "no"))
+                jl_options.banner = 0;
+            else if (!strcmp(optarg, "auto"))
+                jl_options.banner = -1;
             else
-                jl_errorf("julia: invalid argument to --compilecache={yes|no} (%s)", optarg);
+                jl_errorf("julia: invalid argument to --banner={yes|no|auto} (%s)", optarg);
+            break;
+        case opt_sysimage_native_code:
+            if (!strcmp(optarg,"yes"))
+                jl_options.use_sysimage_native_code = JL_OPTIONS_USE_SYSIMAGE_NATIVE_CODE_YES;
+            else if (!strcmp(optarg,"no"))
+                jl_options.use_sysimage_native_code = JL_OPTIONS_USE_SYSIMAGE_NATIVE_CODE_NO;
+            else
+                jl_errorf("julia: invalid argument to --sysimage-native-code={yes|no} (%s)", optarg);
+            break;
+        case opt_compiled_modules:
+            if (!strcmp(optarg,"yes"))
+                jl_options.use_compiled_modules = JL_OPTIONS_USE_COMPILED_MODULES_YES;
+            else if (!strcmp(optarg,"no"))
+                jl_options.use_compiled_modules = JL_OPTIONS_USE_COMPILED_MODULES_NO;
+            else
+                jl_errorf("julia: invalid argument to --compiled-modules={yes|no} (%s)", optarg);
             break;
         case 'C': // cpu-target
             jl_options.cpu_target = strdup(optarg);
+            if (!jl_options.cpu_target)
+                jl_error("julia: failed to allocate memory");
             break;
         case 'p': // procs
             errno = 0;
             if (!strcmp(optarg,"auto")) {
-                jl_options.nprocs = jl_cpu_cores();
+                jl_options.nprocs = jl_cpu_threads();
             }
             else {
                 long nprocs = strtol(optarg, &endptr, 10);
@@ -330,16 +393,23 @@ restart_switch:
                 jl_options.nprocs = (int)nprocs;
             }
             break;
-        case opt_machinefile:
-            jl_options.machinefile = strdup(optarg);
+        case opt_machine_file:
+            jl_options.machine_file = strdup(optarg);
+            if (!jl_options.machine_file)
+                jl_error("julia: failed to allocate memory");
+            break;
+        case opt_project:
+            jl_options.project = optarg ? strdup(optarg) : "@.";
             break;
         case opt_color:
-            if (!strcmp(optarg,"yes"))
+            if (!strcmp(optarg, "yes"))
                 jl_options.color = JL_OPTIONS_COLOR_ON;
-            else if (!strcmp(optarg,"no"))
+            else if (!strcmp(optarg, "no"))
                 jl_options.color = JL_OPTIONS_COLOR_OFF;
+            else if (!strcmp(optarg, "auto"))
+                jl_options.color = JL_OPTIONS_COLOR_AUTO;
             else
-                jl_errorf("julia: invalid argument to --color={yes|no} (%s)", optarg);
+                jl_errorf("julia: invalid argument to --color={yes|no|auto} (%s)", optarg);
             break;
         case opt_history_file:
             if (!strcmp(optarg,"yes"))
@@ -371,12 +441,18 @@ restart_switch:
             break;
         case opt_code_coverage:
             if (optarg != NULL) {
-                if (!strcmp(optarg,"user"))
+                size_t endof = strlen(optarg);
+                if (!strcmp(optarg, "user"))
                     codecov = JL_LOG_USER;
-                else if (!strcmp(optarg,"all"))
+                else if (!strcmp(optarg, "all"))
                     codecov = JL_LOG_ALL;
-                else if (!strcmp(optarg,"none"))
+                else if (!strcmp(optarg, "none"))
                     codecov = JL_LOG_NONE;
+                else if (endof > 5 && !strcmp(optarg + endof - 5, ".info")) {
+                    if (codecov == JL_LOG_NONE)
+                        codecov = JL_LOG_ALL;
+                    jl_options.output_code_coverage = optarg;
+                }
                 else
                     jl_errorf("julia: invalid argument to --code-coverage (%s)", optarg);
                 break;
@@ -434,6 +510,13 @@ restart_switch:
             jl_options.outputbc = optarg;
             if (!jl_options.image_file_specified) jl_options.image_file = NULL;
             break;
+        case opt_output_jit_bc:
+            jl_options.outputjitbc = optarg;
+            break;
+        case opt_output_unopt_bc:
+            jl_options.outputunoptbc = optarg;
+            if (!jl_options.image_file_specified) jl_options.image_file = NULL;
+            break;
         case opt_output_o:
             jl_options.outputo = optarg;
             if (!jl_options.image_file_specified) jl_options.image_file = NULL;
@@ -460,6 +543,14 @@ restart_switch:
             else
                 jl_errorf("julia: invalid argument to --depwarn={yes|no|error} (%s)", optarg);
             break;
+        case opt_warn_overwrite:
+            if (!strcmp(optarg,"yes"))
+                jl_options.warn_overwrite = JL_OPTIONS_WARN_OVERWRITE_ON;
+            else if (!strcmp(optarg,"no"))
+                jl_options.warn_overwrite = JL_OPTIONS_WARN_OVERWRITE_OFF;
+            else
+                jl_errorf("julia: invalid argument to --warn-overwrite={yes|no|} (%s)", optarg);
+            break;
         case opt_inline:
             if (!strcmp(optarg,"yes"))
                 jl_options.can_inline = 1;
@@ -478,6 +569,11 @@ restart_switch:
                 jl_errorf("julia: invalid argument to --polly (%s)", optarg);
             }
             break;
+         case opt_trace_compile:
+            jl_options.trace_compile = strdup(optarg);
+            if (!jl_options.trace_compile)
+                jl_errorf("fatal error: failed to allocate memory: %s", strerror(errno));
+            break;
         case opt_math_mode:
             if (!strcmp(optarg,"ieee"))
                 jl_options.fast_math = JL_OPTIONS_FAST_MATH_OFF;
@@ -489,10 +585,17 @@ restart_switch:
                 jl_errorf("julia: invalid argument to --math-mode (%s)", optarg);
             break;
         case opt_worker:
-            jl_options.worker = strdup(optarg);
+            jl_options.worker = 1;
+            if (optarg != NULL) {
+                jl_options.cookie = strdup(optarg);
+                if (!jl_options.cookie)
+                    jl_error("julia: failed to allocate memory");
+            }
             break;
         case opt_bind_to:
             jl_options.bindto = strdup(optarg);
+            if (!jl_options.bindto)
+                jl_error("julia: failed to allocate memory");
             break;
         case opt_handle_signals:
             if (!strcmp(optarg,"yes"))
@@ -532,3 +635,8 @@ JL_DLLEXPORT void jl_set_ARGS(int argc, char **argv)
         }
     }
 }
+
+JL_DLLEXPORT ssize_t jl_sizeof_jl_options(void)
+{
+    return sizeof(jl_options_t);
+}
