@@ -3,9 +3,9 @@
 /*
   modules and top-level bindings
 */
-#include <assert.h>
 #include "julia.h"
 #include "julia_internal.h"
+#include "julia_assert.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -20,6 +20,7 @@ extern jl_function_t *jl_append_any_func;
 JL_DLLEXPORT jl_module_t *jl_new_module(jl_sym_t *name)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
+    const jl_uuid_t uuid_zero = {0, 0};
     jl_module_t *m = (jl_module_t*)jl_gc_alloc(ptls, sizeof(jl_module_t),
                                                jl_module_type);
     JL_GC_PUSH1(&m);
@@ -27,12 +28,14 @@ JL_DLLEXPORT jl_module_t *jl_new_module(jl_sym_t *name)
     m->name = name;
     m->parent = NULL;
     m->istopmod = 0;
+    m->uuid = uuid_zero;
     static unsigned int mcounter; // simple counter backup, in case hrtime is not incrementing
-    m->uuid = jl_hrtime() + (++mcounter);
-    if (!m->uuid)
-        m->uuid++; // uuid 0 is invalid
+    m->build_id = jl_hrtime() + (++mcounter);
+    if (!m->build_id)
+        m->build_id++; // build id 0 is invalid
     m->primary_world = 0;
     m->counter = 0;
+    m->nospecialize = 0;
     htable_new(&m->bindings, 0);
     arraylist_new(&m->usings, 0);
     if (jl_core_module) {
@@ -52,21 +55,28 @@ uint32_t jl_module_next_counter(jl_module_t *m)
 
 JL_DLLEXPORT jl_value_t *jl_f_new_module(jl_sym_t *name, uint8_t std_imports)
 {
+    // TODO: should we prohibit this during incremental compilation?
     jl_module_t *m = jl_new_module(name);
     JL_GC_PUSH1(&m);
-    m->parent = jl_main_module;
+    m->parent = jl_main_module; // TODO: this is a lie
     jl_gc_wb(m, m->parent);
-    if (std_imports) jl_add_standard_imports(m);
+    if (std_imports)
+        jl_add_standard_imports(m);
     JL_GC_POP();
+    // TODO: should we somehow try to gc-root this correctly?
     return (jl_value_t*)m;
 }
 
-JL_DLLEXPORT void jl_set_istopmod(uint8_t isprimary)
+JL_DLLEXPORT void jl_set_module_nospecialize(jl_module_t *self, int on)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
-    ptls->current_module->istopmod = 1;
+    self->nospecialize = (on ? -1 : 0);
+}
+
+JL_DLLEXPORT void jl_set_istopmod(jl_module_t *self, uint8_t isprimary)
+{
+    self->istopmod = 1;
     if (isprimary) {
-        jl_top_module = ptls->current_module;
+        jl_top_module = self;
         jl_append_any_func = NULL;
     }
 }
@@ -93,24 +103,22 @@ static jl_binding_t *new_binding(jl_sym_t *name)
 }
 
 // get binding for assignment
-JL_DLLEXPORT jl_binding_t *jl_get_binding_wr(jl_module_t *m, jl_sym_t *var)
+JL_DLLEXPORT jl_binding_t *jl_get_binding_wr(jl_module_t *m, jl_sym_t *var, int error)
 {
     jl_binding_t **bp = (jl_binding_t**)ptrhash_bp(&m->bindings, var);
-    jl_binding_t *b;
+    jl_binding_t *b = *bp;
 
-    if (*bp != HT_NOTFOUND) {
-        if ((*bp)->owner == NULL) {
-            (*bp)->owner = m;
-            return *bp;
-        }
-        else if ((*bp)->owner != m) {
-            // TODO: change this to an error soon
-            jl_printf(JL_STDERR,
-                      "WARNING: imported binding for %s overwritten in module %s\n", jl_symbol_name(var), jl_symbol_name(m->name));
-        }
-        else {
-            return *bp;
+    if (b != HT_NOTFOUND) {
+        if (b->owner != m) {
+            if (b->owner == NULL) {
+                b->owner = m;
+            }
+            else if (error) {
+                jl_errorf("cannot assign variable %s.%s from module %s",
+                          jl_symbol_name(b->owner->name), jl_symbol_name(var), jl_symbol_name(m->name));
+            }
         }
+        return *bp;
     }
 
     b = new_binding(var);
@@ -120,6 +128,23 @@ JL_DLLEXPORT jl_binding_t *jl_get_binding_wr(jl_module_t *m, jl_sym_t *var)
     return *bp;
 }
 
+// Hash tables don't generically root their contents, but they do for bindings.
+// Express this to the analyzer.
+#ifdef __clang_analyzer__
+jl_binding_t *_jl_get_module_binding(jl_module_t *m JL_PROPAGATES_ROOT, jl_sym_t *var) JL_NOTSAFEPOINT;
+jl_binding_t **_jl_get_module_binding_bp(jl_module_t *m JL_PROPAGATES_ROOT, jl_sym_t *var) JL_NOTSAFEPOINT;
+#else
+static inline jl_binding_t *_jl_get_module_binding(jl_module_t *m JL_PROPAGATES_ROOT, jl_sym_t *var) JL_NOTSAFEPOINT
+{
+    return (jl_binding_t*)ptrhash_get(&m->bindings, var);
+}
+static inline jl_binding_t **_jl_get_module_binding_bp(jl_module_t *m JL_PROPAGATES_ROOT, jl_sym_t *var) JL_NOTSAFEPOINT
+{
+    return (jl_binding_t**)ptrhash_bp(&m->bindings, var);
+}
+#endif
+
+
 // return module of binding
 JL_DLLEXPORT jl_module_t *jl_get_module_of_binding(jl_module_t *m, jl_sym_t *var)
 {
@@ -130,26 +155,30 @@ JL_DLLEXPORT jl_module_t *jl_get_module_of_binding(jl_module_t *m, jl_sym_t *var
 }
 
 // get binding for adding a method
-// like jl_get_binding_wr, but uses existing imports instead of warning
-// and overwriting.
+// like jl_get_binding_wr, but has different error paths
 JL_DLLEXPORT jl_binding_t *jl_get_binding_for_method_def(jl_module_t *m, jl_sym_t *var)
 {
-    jl_binding_t **bp = (jl_binding_t**)ptrhash_bp(&m->bindings, var);
+    jl_binding_t **bp = _jl_get_module_binding_bp(m, var);
     jl_binding_t *b = *bp;
 
     if (b != HT_NOTFOUND) {
-        if (b->owner != m && b->owner != NULL) {
-            jl_binding_t *b2 = jl_get_binding(b->owner, var);
-            if (b2 == NULL)
-                jl_errorf("invalid method definition: imported function %s.%s does not exist", jl_symbol_name(b->owner->name), jl_symbol_name(var));
-            // TODO: we might want to require explicitly importing types to add constructors
-            if (!b->imported && (b2->value == NULL || !jl_is_type(b2->value))) {
-                jl_errorf("error in method definition: function %s.%s must be explicitly imported to be extended", jl_symbol_name(b->owner->name),
-                          jl_symbol_name(var));
+        if (b->owner != m) {
+            if (b->owner == NULL) {
+                b->owner = m;
+            }
+            else {
+                jl_binding_t *b2 = jl_get_binding(b->owner, var);
+                if (b2 == NULL || b2->value == NULL)
+                    jl_errorf("invalid method definition: imported function %s.%s does not exist",
+                              jl_symbol_name(b->owner->name), jl_symbol_name(var));
+                // TODO: we might want to require explicitly importing types to add constructors
+                if (!b->imported && !jl_is_type(b2->value)) {
+                    jl_errorf("error in method definition: function %s.%s must be explicitly imported to be extended",
+                              jl_symbol_name(b->owner->name), jl_symbol_name(var));
+                }
+                return b2;
             }
-            return b2;
         }
-        b->owner = m;
         return b;
     }
 
@@ -168,6 +197,44 @@ typedef struct _modstack_t {
     struct _modstack_t *prev;
 } modstack_t;
 
+static jl_binding_t *jl_get_binding_(jl_module_t *m JL_PROPAGATES_ROOT, jl_sym_t *var, modstack_t *st);
+
+// find a binding from a module's `usings` list
+static jl_binding_t *using_resolve_binding(jl_module_t *m JL_PROPAGATES_ROOT, jl_sym_t *var, modstack_t *st, int warn)
+{
+    jl_binding_t *b = NULL;
+    jl_module_t *owner = NULL;
+    for(int i=(int)m->usings.len-1; i >= 0; --i) {
+        jl_module_t *imp = (jl_module_t*)m->usings.items[i];
+        jl_binding_t *tempb = _jl_get_module_binding(imp, var);
+        if (tempb != HT_NOTFOUND && tempb->exportp) {
+            tempb = jl_get_binding_(imp, var, st);
+            if (tempb == NULL || tempb->owner == NULL)
+                // couldn't resolve; try next using (see issue #6105)
+                continue;
+            if (owner != NULL && tempb->owner != b->owner &&
+                !tempb->deprecated && !b->deprecated &&
+                !(tempb->constp && tempb->value && b->constp && b->value == tempb->value)) {
+                if (warn) {
+                    jl_printf(JL_STDERR,
+                              "WARNING: both %s and %s export \"%s\"; uses of it in module %s must be qualified\n",
+                              jl_symbol_name(owner->name),
+                              jl_symbol_name(imp->name), jl_symbol_name(var),
+                              jl_symbol_name(m->name));
+                    // mark this binding resolved, to avoid repeating the warning
+                    (void)jl_get_binding_wr(m, var, 0);
+                }
+                return NULL;
+            }
+            if (owner == NULL || !tempb->deprecated) {
+                owner = imp;
+                b = tempb;
+            }
+        }
+    }
+    return b;
+}
+
 // get binding for reading. might return NULL for unbound.
 static jl_binding_t *jl_get_binding_(jl_module_t *m, jl_sym_t *var, modstack_t *st)
 {
@@ -180,35 +247,10 @@ static jl_binding_t *jl_get_binding_(jl_module_t *m, jl_sym_t *var, modstack_t *
         }
         tmp = tmp->prev;
     }
-    jl_binding_t *b = (jl_binding_t*)ptrhash_get(&m->bindings, var);
+    jl_binding_t *b = _jl_get_module_binding(m, var);
     if (b == HT_NOTFOUND || b->owner == NULL) {
-        jl_module_t *owner = NULL;
-        for(int i=(int)m->usings.len-1; i >= 0; --i) {
-            jl_module_t *imp = (jl_module_t*)m->usings.items[i];
-            jl_binding_t *tempb = (jl_binding_t*)ptrhash_get(&imp->bindings, var);
-            if (tempb != HT_NOTFOUND && tempb->exportp) {
-                tempb = jl_get_binding_(imp, var, &top);
-                if (tempb == NULL || tempb->owner == NULL)
-                    // couldn't resolve; try next using (see issue #6105)
-                    continue;
-                if (owner != NULL && tempb->owner != b->owner &&
-                    !(tempb->constp && tempb->value && b->constp && b->value == tempb->value)) {
-                    jl_printf(JL_STDERR,
-                              "WARNING: both %s and %s export \"%s\"; uses of it in module %s must be qualified\n",
-                              jl_symbol_name(owner->name),
-                              jl_symbol_name(imp->name), jl_symbol_name(var),
-                              jl_symbol_name(m->name));
-                    // mark this binding resolved, to avoid repeating the warning
-                    (void)jl_get_binding_wr(m, var);
-                    return NULL;
-                }
-                if (owner == NULL || !tempb->deprecated) {
-                    owner = imp;
-                    b = tempb;
-                }
-            }
-        }
-        if (owner != NULL) {
+        b = using_resolve_binding(m, var, &top, 1);
+        if (b != NULL) {
             // do a full import to prevent the result of this lookup
             // from changing, for example if this var is assigned to
             // later.
@@ -222,20 +264,32 @@ static jl_binding_t *jl_get_binding_(jl_module_t *m, jl_sym_t *var, modstack_t *
     return b;
 }
 
+// get owner of binding when accessing m.var, without resolving the binding
+JL_DLLEXPORT jl_value_t *jl_binding_owner(jl_module_t *m, jl_sym_t *var)
+{
+    jl_binding_t *b = (jl_binding_t*)ptrhash_get(&m->bindings, var);
+    if (b == HT_NOTFOUND || b->owner == NULL) {
+        b = using_resolve_binding(m, var, NULL, 0);
+        if (b == NULL || b->owner == NULL)
+            return jl_nothing;
+    }
+    return (jl_value_t*)b->owner;
+}
+
 JL_DLLEXPORT jl_binding_t *jl_get_binding(jl_module_t *m, jl_sym_t *var)
 {
     return jl_get_binding_(m, var, NULL);
 }
 
-void jl_binding_deprecation_warning(jl_binding_t *b);
+void jl_binding_deprecation_warning(jl_module_t *m, jl_binding_t *b);
 
 JL_DLLEXPORT jl_binding_t *jl_get_binding_or_error(jl_module_t *m, jl_sym_t *var)
 {
-    jl_binding_t *b = jl_get_binding_(m, var, NULL);
+    jl_binding_t *b = jl_get_binding(m, var);
     if (b == NULL)
         jl_undefined_var_error(var);
     if (b->deprecated)
-        jl_binding_deprecation_warning(b);
+        jl_binding_deprecation_warning(m, b);
     return b;
 }
 
@@ -271,8 +325,6 @@ JL_DLLEXPORT int jl_is_imported(jl_module_t *m, jl_sym_t *s)
 static void module_import_(jl_module_t *to, jl_module_t *from, jl_sym_t *s,
                            int explici)
 {
-    if (to == from)
-        return;
     jl_binding_t *b = jl_get_binding(from, s);
     if (b == NULL) {
         jl_printf(JL_STDERR,
@@ -281,6 +333,22 @@ static void module_import_(jl_module_t *to, jl_module_t *from, jl_sym_t *s,
                   jl_symbol_name(to->name));
     }
     else {
+        if (b->deprecated) {
+            if (b->value == jl_nothing) {
+                return;
+            }
+            else if (to != jl_main_module && to != jl_base_module &&
+                     jl_options.depwarn != JL_OPTIONS_DEPWARN_OFF) {
+                /* with #22763, external packages wanting to replace
+                   deprecated Base bindings should simply export the new
+                   binding */
+                jl_printf(JL_STDERR,
+                          "WARNING: importing deprecated binding %s.%s into %s.\n",
+                          jl_symbol_name(from->name), jl_symbol_name(s),
+                          jl_symbol_name(to->name));
+            }
+        }
+
         jl_binding_t **bp = (jl_binding_t**)ptrhash_bp(&to->bindings, s);
         jl_binding_t *bto = *bp;
         if (bto != HT_NOTFOUND) {
@@ -343,18 +411,6 @@ JL_DLLEXPORT void jl_module_use(jl_module_t *to, jl_module_t *from, jl_sym_t *s)
     module_import_(to, from, s, 0);
 }
 
-JL_DLLEXPORT void jl_module_importall(jl_module_t *to, jl_module_t *from)
-{
-    void **table = from->bindings.table;
-    for(size_t i=1; i < from->bindings.size; i+=2) {
-        if (table[i] != HT_NOTFOUND) {
-            jl_binding_t *b = (jl_binding_t*)table[i];
-            if (b->exportp && (b->owner==from || b->imported))
-                jl_module_import(to, from, b->name);
-        }
-    }
-}
-
 JL_DLLEXPORT void jl_module_using(jl_module_t *to, jl_module_t *from)
 {
     if (to == from)
@@ -417,38 +473,44 @@ JL_DLLEXPORT int jl_defines_or_exports_p(jl_module_t *m, jl_sym_t *var)
     return b != HT_NOTFOUND && (b->exportp || b->owner==m);
 }
 
-JL_DLLEXPORT int jl_module_exports_p(jl_module_t *m, jl_sym_t *var)
+JL_DLLEXPORT int jl_module_exports_p(jl_module_t *m, jl_sym_t *var) JL_NOTSAFEPOINT
 {
-    jl_binding_t *b = (jl_binding_t*)ptrhash_get(&m->bindings, var);
+    jl_binding_t *b = _jl_get_module_binding(m, var);
     return b != HT_NOTFOUND && b->exportp;
 }
 
-JL_DLLEXPORT int jl_binding_resolved_p(jl_module_t *m, jl_sym_t *var)
+JL_DLLEXPORT int jl_binding_resolved_p(jl_module_t *m, jl_sym_t *var) JL_NOTSAFEPOINT
 {
-    jl_binding_t *b = (jl_binding_t*)ptrhash_get(&m->bindings, var);
+    jl_binding_t *b = _jl_get_module_binding(m, var);
     return b != HT_NOTFOUND && b->owner != NULL;
 }
 
+JL_DLLEXPORT jl_binding_t *jl_get_module_binding(jl_module_t *m JL_PROPAGATES_ROOT, jl_sym_t *var) JL_NOTSAFEPOINT
+{
+    jl_binding_t *b = _jl_get_module_binding(m, var);
+    return b == HT_NOTFOUND ? NULL : b;
+}
+
 JL_DLLEXPORT jl_value_t *jl_get_global(jl_module_t *m, jl_sym_t *var)
 {
     jl_binding_t *b = jl_get_binding(m, var);
     if (b == NULL) return NULL;
-    if (b->deprecated) jl_binding_deprecation_warning(b);
+    if (b->deprecated) jl_binding_deprecation_warning(m, b);
     return b->value;
 }
 
-JL_DLLEXPORT void jl_set_global(jl_module_t *m, jl_sym_t *var, jl_value_t *val)
+JL_DLLEXPORT void jl_set_global(jl_module_t *m JL_ROOTING_ARGUMENT, jl_sym_t *var, jl_value_t *val JL_ROOTED_ARGUMENT)
 {
-    jl_binding_t *bp = jl_get_binding_wr(m, var);
+    jl_binding_t *bp = jl_get_binding_wr(m, var, 1);
     if (!bp->constp) {
         bp->value = val;
         jl_gc_wb(m, val);
     }
 }
 
-JL_DLLEXPORT void jl_set_const(jl_module_t *m, jl_sym_t *var, jl_value_t *val)
+JL_DLLEXPORT void jl_set_const(jl_module_t *m JL_ROOTING_ARGUMENT, jl_sym_t *var, jl_value_t *val JL_ROOTED_ARGUMENT)
 {
-    jl_binding_t *bp = jl_get_binding_wr(m, var);
+    jl_binding_t *bp = jl_get_binding_wr(m, var, 1);
     if (!bp->constp) {
         bp->value = val;
         bp->constp = 1;
@@ -458,58 +520,103 @@ JL_DLLEXPORT void jl_set_const(jl_module_t *m, jl_sym_t *var, jl_value_t *val)
 
 JL_DLLEXPORT int jl_is_const(jl_module_t *m, jl_sym_t *var)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
-    if (m == NULL) m = ptls->current_module;
     jl_binding_t *b = jl_get_binding(m, var);
     return b && b->constp;
 }
 
-JL_DLLEXPORT void jl_deprecate_binding(jl_module_t *m, jl_sym_t *var)
+// set the deprecated flag for a binding:
+//   0=not deprecated, 1=renamed, 2=moved to another package
+JL_DLLEXPORT void jl_deprecate_binding(jl_module_t *m, jl_sym_t *var, int flag)
 {
     jl_binding_t *b = jl_get_binding(m, var);
-    if (b) b->deprecated = 1;
+    if (b) b->deprecated = flag;
 }
 
 JL_DLLEXPORT int jl_is_binding_deprecated(jl_module_t *m, jl_sym_t *var)
 {
-    jl_binding_t *b = jl_get_binding(m, var);
-    return b && b->deprecated;
+    if (jl_binding_resolved_p(m, var)) {
+        jl_binding_t *b = jl_get_binding(m, var);
+        return b && b->deprecated;
+    }
+    return 0;
 }
 
 extern const char *jl_filename;
 extern int jl_lineno;
 
-void jl_binding_deprecation_warning(jl_binding_t *b)
+char dep_message_prefix[] = "_dep_message_";
+
+jl_binding_t *jl_get_dep_message_binding(jl_module_t *m, jl_binding_t *deprecated_binding)
 {
-    if (b->deprecated && jl_options.depwarn) {
+    size_t prefix_len = strlen(dep_message_prefix);
+    size_t name_len = strlen(jl_symbol_name(deprecated_binding->name));
+    char *dep_binding_name = (char*)alloca(prefix_len+name_len+1);
+    memcpy(dep_binding_name, dep_message_prefix, prefix_len);
+    memcpy(dep_binding_name + prefix_len, jl_symbol_name(deprecated_binding->name), name_len);
+    dep_binding_name[prefix_len+name_len] = '\0';
+    return jl_get_binding(m, jl_symbol(dep_binding_name));
+}
+
+void jl_binding_deprecation_warning(jl_module_t *m, jl_binding_t *b)
+{
+    // Only print a warning for deprecated == 1 (renamed).
+    // For deprecated == 2 (moved to a package) the binding is to a function
+    // that throws an error, so we don't want to print a warning too.
+    if (b->deprecated == 1 && jl_options.depwarn) {
         if (jl_options.depwarn != JL_OPTIONS_DEPWARN_ERROR)
             jl_printf(JL_STDERR, "WARNING: ");
-        if (b->owner)
+        jl_binding_t *dep_message_binding = NULL;
+        if (b->owner) {
             jl_printf(JL_STDERR, "%s.%s is deprecated",
                       jl_symbol_name(b->owner->name), jl_symbol_name(b->name));
-        else
+            dep_message_binding = jl_get_dep_message_binding(b->owner, b);
+        }
+        else {
             jl_printf(JL_STDERR, "%s is deprecated", jl_symbol_name(b->name));
-        jl_value_t *v = b->value;
-        if (v) {
-            if (jl_is_type(v) || jl_is_module(v)) {
-                jl_printf(JL_STDERR, ", use ");
-                jl_static_show(JL_STDERR, v);
-                jl_printf(JL_STDERR, " instead");
+        }
+
+        if (dep_message_binding && dep_message_binding->value) {
+            if (jl_isa(dep_message_binding->value, (jl_value_t*)jl_string_type)) {
+                jl_uv_puts(JL_STDERR, jl_string_data(dep_message_binding->value),
+                    jl_string_len(dep_message_binding->value));
             }
             else {
-                jl_methtable_t *mt = jl_gf_mtable(v);
-                if (mt != NULL && mt->defs.unknown != jl_nothing) {
+                jl_static_show(JL_STDERR, dep_message_binding->value);
+            }
+        }
+        else {
+            jl_value_t *v = b->value;
+            if (v) {
+                if (jl_is_type(v) || jl_is_module(v)) {
                     jl_printf(JL_STDERR, ", use ");
-                    jl_static_show(JL_STDERR, (jl_value_t*)mt->module);
-                    jl_printf(JL_STDERR, ".%s", jl_symbol_name(mt->name));
-                    jl_printf(JL_STDERR, " instead");
+                    jl_static_show(JL_STDERR, v);
+                    jl_printf(JL_STDERR, " instead.");
+                }
+                else {
+                    jl_methtable_t *mt = jl_gf_mtable(v);
+                    if (mt != NULL && (mt->defs != jl_nothing ||
+                                       jl_isa(v, (jl_value_t*)jl_builtin_type))) {
+                        jl_printf(JL_STDERR, ", use ");
+                        if (mt->module != jl_core_module) {
+                            jl_static_show(JL_STDERR, (jl_value_t*)mt->module);
+                            jl_printf(JL_STDERR, ".");
+                        }
+                        jl_printf(JL_STDERR, "%s", jl_symbol_name(mt->name));
+                        jl_printf(JL_STDERR, " instead.");
+                    }
                 }
             }
         }
-        jl_printf(JL_STDERR, ".\n");
+        jl_printf(JL_STDERR, "\n");
 
-        if (jl_options.depwarn != JL_OPTIONS_DEPWARN_ERROR)
-            jl_printf(JL_STDERR, "  likely near %s:%d\n", jl_filename, jl_lineno);
+        if (jl_options.depwarn != JL_OPTIONS_DEPWARN_ERROR) {
+            if (jl_lineno == 0) {
+                jl_printf(JL_STDERR, " in module %s\n", jl_symbol_name(m->name));
+            }
+            else {
+                jl_printf(JL_STDERR, "  likely near %s:%d\n", jl_filename, jl_lineno);
+            }
+        }
 
         if (jl_options.depwarn == JL_OPTIONS_DEPWARN_ERROR) {
             if (b->owner)
@@ -548,19 +655,6 @@ JL_DLLEXPORT void jl_declare_constant(jl_binding_t *b)
     b->constp = 1;
 }
 
-JL_DLLEXPORT jl_value_t *jl_get_current_module(void)
-{
-    jl_ptls_t ptls = jl_get_ptls_states();
-    return (jl_value_t*)ptls->current_module;
-}
-
-JL_DLLEXPORT void jl_set_current_module(jl_value_t *m)
-{
-    jl_ptls_t ptls = jl_get_ptls_states();
-    assert(jl_typeis(m, jl_module_type));
-    ptls->current_module = (jl_module_t*)m;
-}
-
 JL_DLLEXPORT jl_value_t *jl_module_usings(jl_module_t *m)
 {
     jl_array_t *a = jl_alloc_array_1d(jl_array_any_type, 0);
@@ -586,7 +680,7 @@ JL_DLLEXPORT jl_value_t *jl_module_names(jl_module_t *m, int all, int imported)
             int hidden = jl_symbol_name(b->name)[0]=='#';
             if ((b->exportp ||
                  (imported && b->imported) ||
-                 ((b->owner == m) && (all || m == jl_main_module))) &&
+                 (b->owner == m && !b->imported && (all || m == jl_main_module))) &&
                 (all || (!b->deprecated && !hidden))) {
                 jl_array_grow_end(a, 1);
                 //XXX: change to jl_arrayset if array storage allocation for Array{Symbols,1} changes:
@@ -600,9 +694,13 @@ JL_DLLEXPORT jl_value_t *jl_module_names(jl_module_t *m, int all, int imported)
 
 JL_DLLEXPORT jl_sym_t *jl_module_name(jl_module_t *m) { return m->name; }
 JL_DLLEXPORT jl_module_t *jl_module_parent(jl_module_t *m) { return m->parent; }
-JL_DLLEXPORT uint64_t jl_module_uuid(jl_module_t *m) { return m->uuid; }
+JL_DLLEXPORT uint64_t jl_module_build_id(jl_module_t *m) { return m->build_id; }
+JL_DLLEXPORT jl_uuid_t jl_module_uuid(jl_module_t* m) { return m->uuid; }
 
-int jl_is_submodule(jl_module_t *child, jl_module_t *parent)
+// TODO: make this part of the module constructor and read-only?
+JL_DLLEXPORT void jl_set_module_uuid(jl_module_t *m, jl_uuid_t uuid) { m->uuid = uuid; }
+
+int jl_is_submodule(jl_module_t *child, jl_module_t *parent) JL_NOTSAFEPOINT
 {
     while (1) {
         if (parent == child)
